/*******************************************************************************
 * Copyright (c) 2022  Xilinx, Inc. All rights reserved.
 * All rights reserved. This software is made available under
 * the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Xilinx, Inc.
 * Copyright 2022  Xilinx, Inc. All Rights Reserved.
 *******************************************************************************/

package org.eclipse.tcf.internal.dap;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.ImmutableTriple;
import org.eclipse.lsp4j.debug.*;
import org.eclipse.lsp4j.debug.services.IDebugProtocolClient;
import org.eclipse.lsp4j.debug.services.IDebugProtocolServer;
import org.eclipse.lsp4j.jsonrpc.RemoteEndpoint;
import org.eclipse.tcf.commands.*;
import org.eclipse.tcf.core.ChannelTCP;
import org.eclipse.tcf.dap.IRequest;
import org.eclipse.tcf.dap.ITCFObject;
import org.eclipse.tcf.init.EventQueue;
import org.eclipse.tcf.init.TCFSecurityManager;
import org.eclipse.tcf.internal.dap.breakpoint.BreakpointInfo;
import org.eclipse.tcf.internal.dap.breakpoint.TCFBreakpointHandler;
import org.eclipse.tcf.internal.dap.command.ExecuteActionRequest;
import org.eclipse.tcf.internal.dap.command.TerminateCommand;
import org.eclipse.tcf.protocol.Protocol;
import org.eclipse.tcf.services.IExpressions;
import org.eclipse.tcf.services.IRunControl;
import org.eclipse.tcf.util.TCFDataCache;

import java.lang.management.ThreadInfo;
import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.logging.ConsoleHandler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;

import static org.apache.commons.lang3.StringUtils.isBlank;

public class TCFDebugAdapter implements IDebugProtocolServer {
	protected static final Logger LOGGER = Logger.getLogger(TCFDebugAdapter.class.getSimpleName());
	protected IDebugProtocolClient client;
	private TCFLaunch launch;
	private static TCFModelManager model_manager;
	private Map<Integer, ThreadInfo> threadIdNameMap;
	private Map<Integer,String> stackFrameIdContextMap;
	private Map<Integer,String> variableIdContextMap;
	private RemoteEndpoint remoteEndpoint;

	public TCFDebugAdapter() {
		LOGGER.setLevel(Level.SEVERE);
		ConsoleHandler consoleHandler = new ConsoleHandler();
		consoleHandler.setLevel(Level.ALL);
		consoleHandler.setFormatter(new SimpleFormatter());
		LOGGER.addHandler(consoleHandler);
	}

	public void connect(IDebugProtocolClient client) {
		this.client = client;
	}
	public void setRemoteEndpoint(RemoteEndpoint remoteEndpoint) {
		this.remoteEndpoint = remoteEndpoint;
	}


	private void initTcfSettings() {
		Protocol.setEventQueue(new EventQueue());
		ChannelTCP.setSSLContext(TCFSecurityManager.createSSLContext());
	}

	@Override
	public CompletableFuture<Capabilities> initialize(final InitializeRequestArguments args) {
		LOGGER.info(">> initialize " + args);

		initTcfSettings();
		final CompletableFuture<Capabilities> completableFuture = new CompletableFuture<Capabilities>();
		final Capabilities capabilities = new Capabilities();
		capabilities.setSupportsTerminateRequest(true);
		capabilities.setSupportsFunctionBreakpoints(true);
		capabilities.setSupportsDataBreakpoints(true);
		capabilities.setSupportsConditionalBreakpoints(true);
		capabilities.setSupportsHitConditionalBreakpoints(true);
		capabilities.setSupportsSetVariable(true);
		capabilities.setSupportsSetExpression(true);
		capabilities.setSupportsEvaluateForHovers(true);
		completableFuture.complete(capabilities);

		return completableFuture;
	}

	@Override
	public CompletableFuture<Void> attach(final Map<String, Object> args) {
		LOGGER.info(">> attach " + args);
		launchAgent(args);
		return CompletableFuture.completedFuture(null);
	}

	@Override
	public CompletableFuture<Void> launch(final Map<String, Object> args) {
		LOGGER.info(">> launch " + args);
		CompletableFuture<Void> future = launchAgent(args);
		if(future.isCompletedExceptionally()){
			return future;
		}
		return CompletableFuture.completedFuture(null);
	}

	protected CompletableFuture<Void> launchAgent(final Map<String, Object> args) {
		CompletableFuture<Void> future = new CompletableFuture<>();
		try {

			TCFDebugAdapterConfiguration configuration = TCFDebugAdapterConfiguration.fromArgs(args);
			// clientOut("Trigger Agent with " + configuration);

			doLaunch(configuration,client,future);
		} catch (Throwable throwable) {
			LOGGER.log(Level.SEVERE, "Could not launch agent", throwable);
			return CompletableFuture.failedFuture(throwable);
		}
		return future;
	}

	private void doLaunch(TCFDebugAdapterConfiguration configuration, IDebugProtocolClient client, CompletableFuture<Void> future) {
		// Here the TCFModelManager must exist, before the TCFLaunch is created,
		// if not exist, the launchListener in TCFModelManager will not work.
		CompletableFuture<Void> modelManagerInitFuture = new CompletableFuture<>();
		Protocol.invokeAndWait(new Runnable() {
			public void run() {
				if (model_manager == null) {
					model_manager = TCFModelManager.getInstance();
					modelManagerInitFuture.complete(null);
				}
			}
		});
		try {
			modelManagerInitFuture.get();
		} catch (InterruptedException | ExecutionException e) {
			future.completeExceptionally(e);
			return;
		}

		// Send initialized.
		Protocol.invokeLater( new Runnable() {
			public void run() {
				launch = new TCFLaunch(configuration,client,remoteEndpoint);
				launch.doLaunch(future);
			}
		});
		try {
			future.get();
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
			return;
		}
		if(!future.isCompletedExceptionally())
			client.initialized();
	}

	@Override
	public CompletableFuture<EvaluateResponse> evaluate(final EvaluateArguments args) {
		LOGGER.info(">> evaluate " + args.getExpression());
		if(args.getContext().equalsIgnoreCase("hover") &&  args.getFrameId() != null ) {
			String objectId = stackFrameIdContextMap.get(args.getFrameId());
			CompletableFuture<ImmutablePair<String,EvaluateResponse>>  evResponse = new CompletableFuture<>();
			Protocol.invokeAndWait(new Runnable() {
				   @Override
				   public void run() {
					   TCFModel model = model_manager.getModel(launch);
					   Object object = model.getNode(objectId);
					   ImmutablePair<String, EvaluateResponse> hoverExprEval = model.getHoverExprEval(object, args.getExpression(), this);
					   if(hoverExprEval == null ) return;
					   evResponse.complete(hoverExprEval);
				   }
			   }
			);

			ImmutablePair<String, EvaluateResponse> evResponsePair = null;
			try {
				evResponsePair = evResponse.get();
				String anyVarId = evResponsePair.left;
				int variablesReference = evResponsePair.right.getVariablesReference();
				if(!isBlank(evResponsePair.right.getResult())) {
					if (variablesReference > 0) {
						if (variableIdContextMap == null) {
							variableIdContextMap = new HashMap<>();
						}
						variableIdContextMap.put(Math.abs(anyVarId.hashCode()), anyVarId);
					}
					final CompletableFuture<EvaluateResponse> completedFuture = new CompletableFuture<EvaluateResponse>();
					completedFuture.complete(evResponsePair.right);
					return completedFuture;
				}
			} catch (InterruptedException | ExecutionException e) {
				e.printStackTrace();
			}
		}
		if(args.getContext().equalsIgnoreCase("repl")) {
			Protocol.invokeLater(new Runnable() {
				public void run() {
					try {
						String processId = "";
						if(launch.getProcessContext() != null ) {
							processId =  launch.getProcessContext().getID();
						} else {
							if(launch.getContextFilter().stream().findFirst().isPresent()){
								processId =  launch.getContextFilter().stream().findFirst().get();
							}
						}

						byte[] bytes = new byte[0];
						bytes = (args.getExpression() + "\r").getBytes("UTF8");
						launch.writeProcessInputStream(processId,bytes,0,bytes.length);
					} catch (Exception exception) {
						exception.printStackTrace();
					}
				}
			});
			final CompletableFuture<EvaluateResponse> completedFuture = new CompletableFuture<EvaluateResponse>();
			final EvaluateResponse evaluateResponse = new EvaluateResponse();
			completedFuture.complete(evaluateResponse);
			return completedFuture;
		}
		if(args.getContext().equalsIgnoreCase("watch")) {
			List<TCFNodeExecContext> threadExecContexts = null;
			if(args.getFrameId() == null || args.getFrameId()  == 0 ){
				final CompletableFuture<EvaluateResponse> completedFuture = new CompletableFuture<EvaluateResponse>();
				final EvaluateResponse evaluateResponse = new EvaluateResponse();
				evaluateResponse.setResult("not available");
				completedFuture.complete(evaluateResponse);
				return completedFuture;
			}
			String stackFrameContextId = stackFrameIdContextMap.get(args.getFrameId());

			try {
				threadExecContexts = getThreadExecContexts();
				List<TCFNodeExecContext> finalThreadExecContexts = threadExecContexts;
				Protocol.invokeAndWait(new Runnable() {
					   @Override
					   public void run() {
						   TCFModel model = model_manager.getModel(launch);
						   model.getExpressionManager().addExpression(args.getExpression());
						   TCFNodeStackFrame nodeStackFrame = (TCFNodeStackFrame)model.getNode(stackFrameContextId);
						   nodeStackFrame.onExpressionAddedOrRemoved();
					   }
				   }
				);
				CompletableFuture<ImmutablePair<String,EvaluateResponse>>  evResponse = new CompletableFuture<>();
				Protocol.invokeAndWait(new Runnable() {
					   @Override
					   public void run() {
						   TCFModel model = model_manager.getModel(launch);
						   model.getExpressionManager().addExpression(args.getExpression());
						   TCFNodeStackFrame nodeStackFrame = (TCFNodeStackFrame)model.getNode(stackFrameContextId);

						   ImmutablePair<String, EvaluateResponse> exprEvalPair = nodeStackFrame.handleWatchExpr(args.getExpression(), TCFModel.ID_EXPRESSIONS, this);
						   if(exprEvalPair == null ) return;
						   evResponse.complete(exprEvalPair);
					   }
				   }
				);
				ImmutablePair<String, EvaluateResponse> evResponsePair = evResponse.get();
				if(evResponsePair.right.getResult() == null ) {
					final CompletableFuture<EvaluateResponse> completedFuture = new CompletableFuture<EvaluateResponse>();
					final EvaluateResponse evaluateResponse = new EvaluateResponse();
					evaluateResponse.setResult("not available");
					completedFuture.complete(evaluateResponse);
					return completedFuture;
				}
				String anyVarId = evResponsePair.left;
				int variablesReference = evResponsePair.right.getVariablesReference();
				if(!isBlank(evResponsePair.right.getResult())) {
					if (variablesReference > 0) {
						if (variableIdContextMap == null) {
							variableIdContextMap = new HashMap<>();
						}
						variableIdContextMap.put(Math.abs(anyVarId.hashCode()), anyVarId);
					}
					final CompletableFuture<EvaluateResponse> completedFuture = new CompletableFuture<EvaluateResponse>();
					completedFuture.complete(evResponsePair.right);
					return completedFuture;
				}
			} catch (ExecutionException | InterruptedException e) {
				e.printStackTrace();
			}
		}
		final CompletableFuture<EvaluateResponse> completedFuture = new CompletableFuture<EvaluateResponse>();
//		final EvaluateResponse evaluateResponse = new EvaluateResponse();
		 // evaluateResponse.setResult(args.getExpression() + " -Value");
		completedFuture.complete(null);
		 // completedFuture.completeExceptionally(new UnsupportedOperationException("Not supported"));
//		completedFuture.complete(evaluateResponse);
		return completedFuture;
	}

	@Override
	public CompletableFuture<SetExpressionResponse> setExpression(final SetExpressionArguments args) {
		LOGGER.info(">> setExpression " + args.getExpression() + " = " + args.getValue());

		String objectId = stackFrameIdContextMap.get(args.getFrameId());
		CompletableFuture<ImmutablePair<String,EvaluateResponse>>  evResponse = new CompletableFuture<>();
		CompletableFuture<Boolean> setFuture = new CompletableFuture<>();
		Protocol.invokeAndWait(new Runnable() {
			   @Override
			   public void run() {
				   TCFModel model = model_manager.getModel(launch);
				   Object object = model.getNode(objectId);
				   Boolean status = model.setExprValue(object, args.getExpression(), args.getValue(),this,setFuture);
			   }
		   }
		);
		try {
			if(!setFuture.get()){
				final CompletableFuture<SetExpressionResponse> completedFuture = new CompletableFuture<SetExpressionResponse>();
				final SetExpressionResponse setExprResp = new SetExpressionResponse();
				completedFuture.complete(setExprResp);
				return completedFuture;
			}
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}
		EvaluateArguments evArgs = new EvaluateArguments();
		evArgs.setFrameId(args.getFrameId());
		evArgs.setExpression(args.getExpression());
		evArgs.setContext("hover");
		CompletableFuture<EvaluateResponse> evFuture = evaluate(evArgs);
		EvaluateResponse evResp = null;
		try {
			evResp = evFuture.get();
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}
		String result = evResp.getResult();

		final CompletableFuture<SetExpressionResponse> completedFuture = new CompletableFuture<SetExpressionResponse>();
		final SetExpressionResponse setExprResp = new SetExpressionResponse();
		if(evResp.getVariablesReference() != 0 )
			setExprResp.setVariablesReference(evResp.getVariablesReference());
		setExprResp.setNamedVariables(evResp.getNamedVariables());
		setExprResp.setValue(evResp.getResult());
		setExprResp.setType(evResp.getType());
		completedFuture.complete(setExprResp);
		return completedFuture;
	}

	@Override
	public CompletableFuture<Void> disconnect(final DisconnectArguments args) {
		LOGGER.info(">> disconnect");
		return CompletableFuture.completedFuture(null);
	}

	Object getAdaptor(Class adaptorClass) {
		CompletableFuture<Object> adaptorFuture = new CompletableFuture<>();
		Protocol.invokeAndWait(new Runnable() {
				   @Override
				   public void run() {
					   TCFModel model = model_manager.getModel(launch);
					   Object adapter = model.getAdapter(adaptorClass, null);
					   adaptorFuture.complete(adapter);
				   }
			   }
		);
		Object adapter = null;
		try {
			adapter = adaptorFuture.get();
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}
		return adapter;
	}

	@Override
	public CompletableFuture<Void> terminate(final TerminateArguments args) {
		LOGGER.info(">> terminate");
		CompletableFuture<Boolean> future = new CompletableFuture<>();
		ICommandParticipant cmdParticipant = new ICommandParticipant() {
			@Override
			public void requestDone(IRequest request) {
				Protocol.invokeLater(200, new Runnable() {
							@Override
							public void run() {
								launch.getChannel().close();
								TerminatedEventArguments terminatedArgs = new TerminatedEventArguments();
								client.terminated(terminatedArgs);
								future.complete(true);
							}
						}
				);
			}
		};

		Class adapter_class = ITerminateHandler.class;
		Object adaptor = getAdaptor(adapter_class);
		CompletableFuture<TCFNodeExecContext> procExecContextFuture = new CompletableFuture<>();
		Protocol.invokeAndWait(new Runnable() {
			   @Override
			   public void run() {
				   String processId = "";
				   if(launch.getProcessContext() != null ) {
					   processId =  launch.getProcessContext().getID();
				   } else {
						if(launch.getContextFilter() == null ) {
							procExecContextFuture.complete(null);
							return;
						}
					   if(launch.getContextFilter().stream().findFirst().isPresent()){
						   processId =  launch.getContextFilter().stream().findFirst().get();
					   }
				   }

				   TCFModel model = model_manager.getModel(launch);
				   TCFNodeLaunch node = (TCFNodeLaunch) model.getNode("");

				   Map<String, ITCFObject> processNodeExecContexts = node.getChildren().getChildren();
				   String finalProcessId = processId;
				   processNodeExecContexts.entrySet().stream().forEach(e -> {
					   String key = e.getKey();
					   if (!key.equals(finalProcessId)) return;
					   TCFNodeExecContext processNodeExecContext = ((TCFNodeExecContext) e.getValue());
					   procExecContextFuture.complete(processNodeExecContext);
				   });
				   if(!procExecContextFuture.isDone()){
					   procExecContextFuture.complete(null);
				   }
			   }
		   }
		);

		try {
			TCFNodeExecContext procExeContext = procExecContextFuture.get();
			if(procExeContext != null ) {
				ExecuteActionRequest request = new ExecuteActionRequest(new Object[]{procExeContext});
				request.setCommandParticipant(cmdParticipant);
				((TerminateCommand) adaptor).execute(request);
			}
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}

		CompletableFuture<Void> resFuture = new CompletableFuture<Void>();
		resFuture.complete(null);
		return resFuture;
	}

	@Override
	public CompletableFuture<Void> setExceptionBreakpoints(final SetExceptionBreakpointsArguments args) {
		LOGGER.info(">> setExceptionBreakpoints");
		return CompletableFuture.completedFuture(null);
	}

	@Override
	public CompletableFuture<SetBreakpointsResponse> setBreakpoints(final SetBreakpointsArguments args) {
		LOGGER.info(">> setBreakpoints");
		if (args.getSource() == null || args.getSource().getPath() == null) {
			resetForeignBreakpoints(args.getBreakpoints());
			final CompletableFuture<SetBreakpointsResponse> future = new CompletableFuture<>();
			SetBreakpointsResponse response = new SetBreakpointsResponse();
			response.setBreakpoints(new Breakpoint[0]);
			future.complete(response);
			return future;
		} else {
			if(launch == null ) {
				final CompletableFuture<SetBreakpointsResponse> future = new CompletableFuture<>();
				SetBreakpointsResponse response = new SetBreakpointsResponse();
				response.setBreakpoints(new Breakpoint[0]);
				future.complete(response);
				return future;
			}
			TCFBreakpointHandler breakpointHandler = launch.getBreakpointHandler();
			BreakpointInfo[] breakpointInfo = breakpointHandler.setBreakpoints(args);
			Breakpoint[] actualBreakpoints = breakpointHandler.createBreakpointsFromInfo(breakpointInfo);
			updateBreakpointLocations(breakpointInfo);
			final CompletableFuture<SetBreakpointsResponse> future = new CompletableFuture<>();
			SetBreakpointsResponse response = new SetBreakpointsResponse();
			response.setBreakpoints(actualBreakpoints);
			future.complete(response);
			return future;
		}
	}

	private void resetForeignBreakpoints(SourceBreakpoint[] sourceBreakpoints) {
		List<BreakpointInfo> foreignBreakpoints =
				TCFBreakpointsModel.INSTANCE.getForeignBreakpoints(BreakpointInfo.Type.SOURCE, launch.getChannel());
		List<BreakpointInfo> breakpointsToAdd = new ArrayList<>();
		for (BreakpointInfo info : foreignBreakpoints) {
			boolean exists = Arrays.stream(sourceBreakpoints).anyMatch(
					bp -> info.getLine() == bp.getLine() && info.getColumn() == bp.getColumn()
			);
			if (!exists) {
				breakpointsToAdd.add(info);
			}
		}

		Breakpoint[] breakpoints = launch.getBreakpointHandler()
				.createBreakpointsFromInfo(breakpointsToAdd.toArray(new BreakpointInfo[0]));
		if (breakpoints == null || breakpoints.length == 0) {
			return;
		}
		for (Breakpoint breakpoint : breakpoints) {
			BreakpointEventArguments eventArgs = new BreakpointEventArguments();
			eventArgs.setBreakpoint(breakpoint);
			eventArgs.setReason(BreakpointEventArgumentsReason.NEW);
			launch.getClient().breakpoint(eventArgs);
		}
	}

	private void updateBreakpointLocations(BreakpointInfo[] bpInfo) {
		try {
			List<TCFNodeExecContext> threadExecContexts = getThreadExecContexts();
			TCFBreakpointHandler breakpointHandler = launch.getBreakpointHandler();
			TCFBreakpointsStatus bpStatus = launch.getBreakpointsStatus();
			IDebugProtocolClient client = launch.getClient();
			breakpointHandler.fireBreakpointLineUpdates(bpInfo, threadExecContexts, bpStatus, client, false);
		} catch (ExecutionException | InterruptedException e) {
			e.printStackTrace();
		}
	}

	@Override
	public CompletableFuture<SetFunctionBreakpointsResponse> setFunctionBreakpoints(final SetFunctionBreakpointsArguments args) {
		LOGGER.info(">> setFunctionBreakpoints");
		resetForeignBreakpoints(args.getBreakpoints());
		BreakpointInfo[] breakpointInfo = launch.getBreakpointHandler().setFunctionBreakpoints(args);
		Breakpoint[] actualBreakpoints = launch.getBreakpointHandler().createBreakpointsFromInfo(breakpointInfo);
		final CompletableFuture<SetFunctionBreakpointsResponse> future = new CompletableFuture<>();
		SetFunctionBreakpointsResponse response = new SetFunctionBreakpointsResponse();
		response.setBreakpoints(actualBreakpoints);
		future.complete(response);
		return future;
	}

	private void resetForeignBreakpoints(FunctionBreakpoint[] functionBreakpoints) {
		List<BreakpointInfo> foreignBreakpoints =
				TCFBreakpointsModel.INSTANCE.getForeignBreakpoints(BreakpointInfo.Type.FUNCTION, launch.getChannel());
		List<BreakpointInfo> breakpointsToAdd = new ArrayList<>();
		for (BreakpointInfo info : foreignBreakpoints) {
			boolean exists = Arrays.stream(functionBreakpoints).anyMatch(
					bp -> Objects.equals(info.getFunction(), bp.getName())
			);
			if (!exists) {
				breakpointsToAdd.add(info);
			}
		}
		if (breakpointsToAdd.isEmpty()) {
			return;
		}
		try {
			BreakpointInfo[] bpInfo = breakpointsToAdd.toArray(new BreakpointInfo[0]);
			List<TCFNodeExecContext> threadExecContexts = getThreadExecContexts();
			TCFBreakpointHandler breakpointHandler = launch.getBreakpointHandler();
			TCFBreakpointsStatus bpStatus = launch.getBreakpointsStatus();
			IDebugProtocolClient client = launch.getClient();
			breakpointHandler.fireBreakpointLineUpdates(bpInfo, threadExecContexts, bpStatus, client, true);
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	private class ThreadInfo {
		public ThreadInfo(int threadId, String threadName, String threadContextId) {
			this.threadId = threadId;
			this.threadName = threadName;
			this.threadContextId = threadContextId;
		}

		public int getThreadId() {
			return threadId;
		}

		public String getThreadName() {
			return threadName;
		}

		public String getThreadContextId() {
			return threadContextId;
		}

		private int threadId;
		private String threadName;
		private String threadContextId; // TCF Context ID;


	}

	public List<TCFNodeExecContext> getThreadExecContexts() throws ExecutionException, InterruptedException {

		CompletableFuture<List<TCFNodeExecContext>> threadNodeExeContextListFuture = new CompletableFuture<>();
		Protocol.invokeAndWait(new Runnable() {
				 @Override
				 public void run() {

					 String processId = "";
					 if(launch.getProcessContext() != null ) {
					 	processId =  launch.getProcessContext().getID();
					 } else {
						 if(launch.getContextFilter().stream().findFirst().isPresent()){
							 processId =  launch.getContextFilter().stream().findFirst().get();
						 }
					 }
					 TCFModel model = model_manager.getModel(launch);
					 List<TCFNodeExecContext> threadNodeExecContextList = new LinkedList<>();
					 // This is the launch node.
					 if (model != null) {
						 TCFNodeLaunch node = (TCFNodeLaunch) model.getNode("");
						 TCFChildren children = node.getChildren();
						 if (!children.validate(this)) return;
						 if (children.isDisposed()) {
							 threadNodeExeContextListFuture.complete(null);
						 }
						 Map<String, ITCFObject> children1 = children.getChildren();
						 for(Map.Entry<String, ITCFObject> e :children1.entrySet() ){
							 String key = e.getKey();
							 if (!key.equalsIgnoreCase(processId)) continue;
							 TCFNodeExecContext procNodeExecContext = (TCFNodeExecContext) e.getValue();
							 TCFChildren tcfChildren = procNodeExecContext.getChildren();
							 if (!tcfChildren.validate(this)) return;
							 if (tcfChildren.isDisposed()) {
								 threadNodeExeContextListFuture.complete(null);
							 }
							 tcfChildren.getChildren().entrySet().stream().forEach(e1 -> {
								 threadNodeExecContextList.add((TCFNodeExecContext) e1.getValue());
							 });
						 }
						 threadNodeExeContextListFuture.complete(threadNodeExecContextList);
					 }

				 }
			 }
		);
		List<TCFNodeExecContext> tcfNodeExecContexts = threadNodeExeContextListFuture.get();
		return tcfNodeExecContexts;
	}

	@Override
	public CompletableFuture<ThreadsResponse> threads() {
		LOGGER.info(">> threads");
		final CompletableFuture<ThreadsResponse> future = new CompletableFuture<>();
		final ThreadsResponse response = new ThreadsResponse();
		try {
			List<TCFNodeExecContext> threadExecContexts = getThreadExecContexts();
			CompletableFuture<Map<Integer, ThreadInfo>>  threadIdNameMapFuture = new CompletableFuture<>();
				Protocol.invokeLater(new Runnable() {
					@Override
					public void run() {
						Map<Integer, TCFDebugAdapter.ThreadInfo> threadIdNameMap = new TreeMap<>();
						for(TCFNodeExecContext threadNodeExecContext : threadExecContexts){
							TCFDataCache<IRunControl.RunControlContext> runContext = threadNodeExecContext.getRunContext();
							if (!runContext.validate(this)) return;
							if (runContext.isDisposed()) {
								// error handling
								threadIdNameMapFuture.complete(null);
								return;
							}
							Map<String, Object> properties = runContext.getData().getProperties();
							String threadIdInTcf = (String) properties.get("Name");
							String additionalInfo = (String) properties.get("AdditionalInfo");
							String threadName = threadIdInTcf;
							String threadId = threadNodeExecContext.getID();
							if (additionalInfo != null) {
								threadName = threadIdInTcf + " " + additionalInfo;
							}
							threadIdNameMap.put(Integer.parseInt(threadIdInTcf), new TCFDebugAdapter.ThreadInfo(Integer.parseInt(threadIdInTcf), threadName, threadId));
						}

						threadIdNameMapFuture.complete(threadIdNameMap);
					}
				});

			threadIdNameMap = threadIdNameMapFuture.get();
		} catch (ExecutionException | InterruptedException e) {
			e.printStackTrace();
		}
		List<org.eclipse.lsp4j.debug.Thread> threadList = new ArrayList<>();

		threadIdNameMap.entrySet().stream().forEachOrdered(e->{
			org.eclipse.lsp4j.debug.Thread threadEntry = new org.eclipse.lsp4j.debug.Thread();
			threadEntry.setId(e.getKey());
			threadEntry.setName(e.getValue().getThreadName());
			threadList.add(threadEntry);
		});
		if(threadList.size() == 0){
			// System.out.println(" Its empty ");
		}
		response.setThreads(threadList.toArray(new org.eclipse.lsp4j.debug.Thread [0]));
		future.complete(response);
		return future;
	}

	@Override
	public CompletableFuture<StackTraceResponse> stackTrace(final StackTraceArguments args) {
		if (stackFrameIdContextMap == null) {
			stackFrameIdContextMap = new HashMap<>();
		}
		LOGGER.info(">> stackTrace of thread " + args.getThreadId());

		String threadContextId = threadIdNameMap.get(args.getThreadId()).getThreadContextId();
		List<StackFrame> stackFrameList = null;
		try {
			List<TCFNodeExecContext> threadExecContexts = getThreadExecContexts();
			CompletableFuture<List<StackFrame>> sfListFuture = new CompletableFuture<>();
			Protocol.invokeAndWait(new Runnable() {
				   @Override
				   public void run() {
					   List<StackFrame> stackFrameList = new LinkedList<>();
					   TCFModel model = model_manager.getModel(launch);

					   for (TCFNodeExecContext threadNodeExecContext : threadExecContexts) {
						   String contextId = threadNodeExecContext.getID();
						   if (!contextId.equals(threadContextId)) continue;
						   ImmutablePair<StackFrame, String> sfContextIdPair = model.displaySource(threadNodeExecContext, this);
						   if (sfContextIdPair != null) {
							   StackFrame topStackFrame = sfContextIdPair.left;
							   if (topStackFrame != null) {
								   stackFrameList.add(topStackFrame);
								   stackFrameIdContextMap.put(topStackFrame.getId(), sfContextIdPair.right);
							   }
						   } else {
							   return;
						   }
						   Collection<TCFNode> stackFrameNodes = threadNodeExecContext.getStackTrace().getNodes();
						   TCFNode[] orderedStackFrameNodes = new TCFNode[stackFrameNodes.size()];
						   stackFrameNodes.stream().forEachOrdered(tcfNodeEntry -> {
									   String id = tcfNodeEntry.getID();
									   String fpNum = id.split("\\.")[0].replace("FP", "");
									   int integer = Integer.parseInt(fpNum);
									   orderedStackFrameNodes[integer] = tcfNodeEntry;
								   }
						   );

						   for (int i = 1; i < orderedStackFrameNodes.length; i++) {
							   TCFNodeStackFrame tcfNodeStackFrame = (TCFNodeStackFrame) orderedStackFrameNodes[i];
							   ImmutablePair<StackFrame, String> sfImPair2 = model.displaySource(tcfNodeStackFrame, this);
							   if (sfImPair2 != null) {
									if(sfImPair2.left !=null) {
										stackFrameList.add(sfImPair2.left);
										stackFrameIdContextMap.put(sfImPair2.left.getId(), sfImPair2.right);
									}
							   } else {
								   return;
							   }
						   }
					   }
					   sfListFuture.complete(stackFrameList);
				   }
			   }
			);
			stackFrameList = sfListFuture.get();
		} catch (ExecutionException | InterruptedException e) {
			e.printStackTrace();
		}

		final StackTraceResponse response = new StackTraceResponse();

		if (stackFrameList != null && stackFrameList.size() > 0) {
			response.setStackFrames(stackFrameList.toArray(new StackFrame[0]));
			response.setTotalFrames(stackFrameList.size());
		} else {
			final StackFrame stackFrame1 = new StackFrame();
			stackFrame1.setName("No Stack for Thread " + args.getThreadId());
			stackFrame1.setPresentationHint(StackFramePresentationHint.LABEL);
			response.setStackFrames(new StackFrame[]{stackFrame1});
			response.setTotalFrames(1);
		}
		final CompletableFuture<StackTraceResponse> future = new CompletableFuture<>();
		future.complete(response);
		return future;
	}
		@Override
	public CompletableFuture<ScopesResponse> scopes(final ScopesArguments args) {
		LOGGER.info(">> scopes of frame " + args.getFrameId());
		String stackFrameContextId = stackFrameIdContextMap.get(args.getFrameId());

		final Scope scope = new Scope();
		scope.setName("Locals");
		scope.setVariablesReference(args.getFrameId());
		CompletableFuture<ImmutablePair<Integer,Integer>> countFuture = new CompletableFuture<>();
		Protocol.invokeAndWait(new Runnable() {
			   @Override
			   public void run() {
				   TCFModel model = model_manager.getModel(launch);
				   TCFNodeStackFrame nodeStackFrame = (TCFNodeStackFrame)model.getNode(stackFrameContextId);
				   if(nodeStackFrame == null || nodeStackFrame.isDisposed() ){
					   countFuture.complete(null);
				   }
				   ImmutablePair<Integer, Integer> variablesCount = model.getVariablesCount(nodeStackFrame, this);
				   if(variablesCount == null ) return;
				   countFuture.complete(variablesCount);
			   }
		   }
		);
		try {
			ImmutablePair<Integer, Integer> countPair = countFuture.get();
			if(countPair != null ){
				scope.setNamedVariables(countPair.left);
				scope.setIndexedVariables(countPair.right);
			}
			scope.setPresentationHint("locals");
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}
		final CompletableFuture<ScopesResponse> future = new CompletableFuture<>();
		final ScopesResponse response = new ScopesResponse();
		response.setScopes(new Scope[] { scope });
		future.complete(response);
		return future;
	}

	@Override
	public CompletableFuture<VariablesResponse> variables(final VariablesArguments args) {
		LOGGER.info(">> variables of variable reference " + args.getVariablesReference());
		if(variableIdContextMap == null ){
			variableIdContextMap = new HashMap<>();
		}

		String objectId = stackFrameIdContextMap.get(args.getVariablesReference());
		if( objectId == null ){
			objectId = variableIdContextMap.get(args.getVariablesReference());
		}

		String finalObjectId = objectId;
		CompletableFuture<List<ImmutableTriple<String,Variable,Boolean>>> variablesFuture = new CompletableFuture<>();
		Protocol.invokeLater(new Runnable() {
			   @Override
			   public void run() {
				   TCFModel model = model_manager.getModel(launch);
				   Object object = model.getNode(finalObjectId);
				   List<ImmutableTriple<String, Variable,Boolean>> variables = model.getVariables(object, args.getFilter(), this);
				   if(variables == null ) return;
				   variablesFuture.complete(variables);
			   }
		   }
		);

		final CompletableFuture<VariablesResponse> future = new CompletableFuture<>();
		final VariablesResponse response = new VariablesResponse();
		List<ImmutableTriple<String,Variable,Boolean>> variables = null;
		try {
			variables = variablesFuture.get();
			if(variables == null ){
				response.setVariables(new Variable[0]);
			}else {
				variables.forEach(e->{
					if(e.right) {
						variableIdContextMap.put(Math.abs(e.left.hashCode()),e.left);
					}
				});
				response.setVariables(variables.stream().map(e -> e.middle).toArray(Variable[]::new));
			}
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}
		future.complete(response);
		return future;
	}

	@Override
	public CompletableFuture<Void> pause(PauseArguments args) {
		LOGGER.info(">> pause " + args);

		String threadContextId = threadIdNameMap.get(args.getThreadId()).getThreadContextId();
		CompletableFuture<Boolean> future = new CompletableFuture<>();
		ICommandParticipant cmdParticipant = new ICommandParticipant() {
			@Override
			public void requestDone(IRequest request) {
				StoppedEventArguments stoppedArgs = new StoppedEventArguments();
				stoppedArgs.setDescription("step");
				stoppedArgs.setReason("step");
				stoppedArgs.setThreadId(args.getThreadId());
				//stoppedArgs.setAllThreadsStopped(false);
				client.stopped(stoppedArgs);
				future.complete(true);
			}
		};

		Class adapter_class = ISuspendHandler.class;
		executeRunControlAction(cmdParticipant,adapter_class,threadContextId);

		StoppedEventArguments stopped = new StoppedEventArguments();
		stopped.setReason(StoppedEventArgumentsReason.PAUSE);
		stopped.setThreadId(args.getThreadId());
		this.client.stopped(stopped);
		return CompletableFuture.completedFuture(null);
	}

	@Override
	public CompletableFuture<ContinueResponse> continue_(ContinueArguments args) {
		LOGGER.info(">> continue " + args);

		String threadContextId = threadIdNameMap.get(args.getThreadId()).getThreadContextId();
		CompletableFuture<Boolean> future = new CompletableFuture<>();
		ICommandParticipant cmdParticipant = new ICommandParticipant() {
			@Override
			public void requestDone(IRequest request) {
				ContinuedEventArguments continueArgs = new ContinuedEventArguments();
				continueArgs.setThreadId(args.getThreadId());
				continueArgs.setAllThreadsContinued(false);
				// stoppedArgs.setAllThreadsStopped(true);
				client.continued(continueArgs);
				future.complete(true);
			}
		};
		Class adapter_class = IResumeHandler.class;
		executeRunControlAction(cmdParticipant,adapter_class,threadContextId);

		ContinueResponse response = new ContinueResponse();
		return CompletableFuture.completedFuture(response);
	}

	private void executeRunControlAction(ICommandParticipant cmdParticipant, Class adapter_class, String threadContextId) {
		try {
			Object adapter = getAdaptor(adapter_class);
			List<TCFNodeExecContext> threadExecContexts = getThreadExecContexts();
			for(TCFNodeExecContext threadExeContext : threadExecContexts){
				if (!threadExeContext.getID().equals(threadContextId)) continue;
				ExecuteActionRequest request = new ExecuteActionRequest(new Object[]{threadExeContext});
				request.setCommandParticipant(cmdParticipant);
				((IDebugCommandHandler)adapter).execute(request);
			}
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}
	}

	/**
	 * UnsupportedOperationException below
	 */

	@Override
	public CompletableFuture<Void> cancel(CancelArguments args) {
		LOGGER.info(">> cancel " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<RunInTerminalResponse> runInTerminal(RunInTerminalRequestArguments args) {
		LOGGER.info(">> runInTerminal " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<Void> configurationDone(ConfigurationDoneArguments args) {
		LOGGER.info(">> configurationDone " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<Void> restart(RestartArguments args) {
		LOGGER.info(">> restart " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<BreakpointLocationsResponse> breakpointLocations(BreakpointLocationsArguments args) {
		LOGGER.info(">> breakpointLocations " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<DataBreakpointInfoResponse> dataBreakpointInfo(DataBreakpointInfoArguments args) {
		LOGGER.info(">> dataBreakpointInfo " + args);
		DataBreakpointInfoResponse response = new DataBreakpointInfoResponse();
		return CompletableFuture.completedFuture(response);
	}

	@Override
	public CompletableFuture<SetDataBreakpointsResponse> setDataBreakpoints(SetDataBreakpointsArguments args) {
		LOGGER.info(">> setDataBreakpoints " + args);
		SetDataBreakpointsResponse response = new SetDataBreakpointsResponse();
		return CompletableFuture.completedFuture(response);
	}

	@Override
	public CompletableFuture<SetInstructionBreakpointsResponse> setInstructionBreakpoints(
			SetInstructionBreakpointsArguments args) {
		LOGGER.info(">> setInstructionBreakpoints " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<Void> next(NextArguments args) {
		LOGGER.info(">> next " + args);
		String threadContextId = threadIdNameMap.get(args.getThreadId()).getThreadContextId();
		CompletableFuture<Boolean> future = new CompletableFuture<>();
		ICommandParticipant cmdParticipant =new ICommandParticipant() {
			@Override
			public void requestDone(IRequest request) {
				StoppedEventArguments stoppedArgs = new StoppedEventArguments();
				stoppedArgs.setDescription("step");
				stoppedArgs.setReason("step");
				stoppedArgs.setThreadId(args.getThreadId());
				//stoppedArgs.setAllThreadsStopped(false);
				client.stopped(stoppedArgs);
				future.complete(true);
			}
		};
		Class adapter_class = IStepOverHandler.class;
		executeRunControlAction(cmdParticipant,adapter_class,threadContextId);

		CompletableFuture<Void> resFuture = new CompletableFuture<Void>();
		resFuture.complete(null);
		return resFuture;
	}

	@Override
	public CompletableFuture<Void> stepIn(StepInArguments args) {
		LOGGER.info(">> stepIn " + args);
		String threadContextId = threadIdNameMap.get(args.getThreadId()).getThreadContextId();
		CompletableFuture<Boolean> future = new CompletableFuture<>();
		ICommandParticipant cmdParticipant =new ICommandParticipant() {
			@Override
			public void requestDone(IRequest request) {
				StoppedEventArguments stoppedArgs = new StoppedEventArguments();
				stoppedArgs.setDescription("step");
				stoppedArgs.setReason("step");
				stoppedArgs.setThreadId(args.getThreadId());
				// stoppedArgs.setAllThreadsStopped(true);
				client.stopped(stoppedArgs);
				future.complete(true);
			}
		};
		Class adapter_class = IStepIntoHandler.class;
		executeRunControlAction(cmdParticipant,adapter_class,threadContextId);

		CompletableFuture<Void> resFuture = new CompletableFuture<Void>();
		resFuture.complete(null);
		return resFuture;

	}

	@Override
	public CompletableFuture<Void> stepOut(StepOutArguments args) {
		LOGGER.info(">> stepOut " + args);
		String threadContextId = threadIdNameMap.get(args.getThreadId()).getThreadContextId();
		CompletableFuture<Boolean> future = new CompletableFuture<>();
		ICommandParticipant cmdParticipant =new ICommandParticipant() {
			@Override
			public void requestDone(IRequest request) {
				StoppedEventArguments stoppedArgs = new StoppedEventArguments();
				stoppedArgs.setDescription("step");
				stoppedArgs.setReason("step");
				stoppedArgs.setThreadId(args.getThreadId());
				// stoppedArgs.setAllThreadsStopped(true);
				client.stopped(stoppedArgs);
				future.complete(true);
			}
		};
		Class adapter_class = IStepReturnHandler.class;
		executeRunControlAction(cmdParticipant,adapter_class,threadContextId);
		CompletableFuture<Void> resFuture = new CompletableFuture<Void>();
		resFuture.complete(null);
		return resFuture;
	}

	@Override
	public CompletableFuture<Void> stepBack(StepBackArguments args) {
		LOGGER.info(">> stepBack " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<Void> reverseContinue(ReverseContinueArguments args) {
		LOGGER.info(">> reverseContinue " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<Void> restartFrame(RestartFrameArguments args) {
		LOGGER.info(">> restartFrame " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<Void> goto_(GotoArguments args) {
		LOGGER.info(">> goto " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<SetVariableResponse> setVariable(SetVariableArguments args) {
		LOGGER.info(">> setVariable " + args);

		String objectId = stackFrameIdContextMap.get(args.getVariablesReference());
		if( objectId == null ){
			objectId = variableIdContextMap.get(args.getVariablesReference());
		}
		String finalObjectId = objectId;
		CompletableFuture<List<ImmutableTriple<String,Variable,Boolean>>> variablesFuture = new CompletableFuture<>();

		Protocol.invokeLater(new Runnable() {
				 @Override
				 public void run() {
					 TCFModel model = model_manager.getModel(launch);
					 Object object = model.getNode(finalObjectId);
					 List<ImmutableTriple<String, Variable,Boolean>> variables = model.getVariables(object, null, this);
					 if(variables == null ) return;
					 variablesFuture.complete(variables);
				 }
			 }
		);
		try {
			List<ImmutableTriple<String, Variable,Boolean>> variables = variablesFuture.get();
			if(variables == null ){
				SetVariableResponse resp = new SetVariableResponse();
				CompletableFuture<SetVariableResponse> retFuture = new CompletableFuture<>();
				retFuture.complete(resp);
				return retFuture;
			}else {
				Optional<ImmutableTriple<String, Variable, Boolean>> first = variables.stream().filter(
						e -> (e.middle.getName().equalsIgnoreCase(args.getName()))).findFirst();
				if(!first.isPresent()){
					SetVariableResponse resp = new SetVariableResponse();
					CompletableFuture<SetVariableResponse> retFuture = new CompletableFuture<>();
					retFuture.complete(resp);
					return retFuture;
				}
				String expressionId = first.get().left;
				CompletableFuture<Boolean> setFuture = new CompletableFuture<>();

				Protocol.invokeLater(new Runnable() {
					@Override
					public void run() {
						TCFModel model = model_manager.getModel(launch);
						TCFNodeExpression nodeExpression = (TCFNodeExpression) model.getNode(expressionId);
						try {
							Boolean status = nodeExpression.modify(args.getValue(), this, setFuture);
						} catch (Exception exception) {
							exception.printStackTrace();
						}
					}
				});
				Boolean isSet = setFuture.get();
				CompletableFuture<String> valueStr = new CompletableFuture<>();
				Protocol.invokeLater(new Runnable() {
					@Override
					public void run() {
						TCFModel model = model_manager.getModel(launch);
						TCFNodeExpression nodeExpression = (TCFNodeExpression) model.getNode(expressionId);
						TCFDataCache<String> varExprText = nodeExpression.getExpressionText();
						if (!varExprText.validate(this)) return;
						String varExprTextData = varExprText.getData();
						TCFDataCache<String> typeName = nodeExpression.getTypeName();
						if (!typeName.validate(this)) return;
						String typeNameData = typeName.getData();
						TCFDataCache<String> baseText = nodeExpression.getBaseText();
						if(!baseText.validate(this)) return;
						String baseTextData = baseText.getData();

						TCFNode.TCFData<IExpressions.Expression> varExpression = nodeExpression.getVarExpression();
						if(!varExpression.validate(this)) return;
						IExpressions.Expression data = varExpression.getData();
						TCFDataCache<IExpressions.Value> exprValue = nodeExpression.getValue();
						if(!exprValue.validate(this)) return;
						IExpressions.Value exprValueData = exprValue.getData();
						TCFDataCache<String> valueString = nodeExpression.getValueString();
						if (!valueString.validate(this)) return;
						String valueStringData = valueString.getData();
						valueStr.complete(valueStringData);
					}
				}
				);
				String newValue = valueStr.get();
				SetVariableResponse resp = new SetVariableResponse();
					Variable variable = first.get().middle;
					resp.setVariablesReference(variable.getVariablesReference());
					resp.setNamedVariables(variable.getNamedVariables());
					resp.setType(variable.getType());
					resp.setValue(valueStr.get());
					// resp.setVariablesReference(args.getVariablesReference());
					CompletableFuture<SetVariableResponse> retFuture = new CompletableFuture<>();
					retFuture.complete(resp);
/*	Used for verifying the modified variable value...

					VariablesArguments varargs = new VariablesArguments();
					varargs.setVariablesReference(args.getVariablesReference());
					CompletableFuture<VariablesResponse> variables2 = variables(varargs);

					try {
						variables2.get();
					} catch (InterruptedException | ExecutionException e) {
						e.printStackTrace();
					}

					for( Map.Entry<Integer,String> entry : stackFrameIdContextMap.entrySet() ){

						if(!entry.getValue().contains("FP0")) continue;
						VariablesArguments varargs2 = new VariablesArguments();
						varargs2.setVariablesReference(entry.getKey());
						CompletableFuture<VariablesResponse> variables3 = variables(varargs2);

						try {
							variables3.get();
						} catch (InterruptedException | ExecutionException e) {
							e.printStackTrace();
						}

						EvaluateArguments evArgs = new EvaluateArguments();
						evArgs.setFrameId(entry.getKey());
						evArgs.setExpression("strvar");
						evArgs.setContext("hover");
						CompletableFuture<EvaluateResponse> evaluate = evaluate(evArgs);
						EvaluateResponse evResp = evaluate.get();
						String result = evResp.getResult();
						break;
					}

 */
					return retFuture;
			}
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}
		SetVariableResponse resp = new SetVariableResponse();
		CompletableFuture<SetVariableResponse> retFuture = new CompletableFuture<>();
		retFuture.complete(resp);
		return retFuture;
	}

	@Override
	public CompletableFuture<SourceResponse> source(SourceArguments args) {
		LOGGER.info(">> source " + args);

		int sourceReference = args.getSourceReference();
		CompletableFuture<SourceResponse> sourceRespFuture = new CompletableFuture<>();
		Protocol.invokeLater(new Runnable() {
								 @Override
								 public void run() {
									 TCFModel model = model_manager.getModel(launch);
									 SourceResponse sourceResp = model.getDisassembly(sourceReference, this);
									 if(sourceResp == null ) return;
									 sourceRespFuture.complete(sourceResp);
								 }
							 }
		);

		CompletableFuture<SourceResponse> respFuture = new CompletableFuture<>();
		try {
			SourceResponse resp = sourceRespFuture.get();
			respFuture.complete(resp);
		} catch (InterruptedException | ExecutionException e) {
			e.printStackTrace();
		}
		if(!respFuture.isDone()) respFuture.complete(null);
		return respFuture;
	}

	@Override
	public CompletableFuture<Void> terminateThreads(TerminateThreadsArguments args) {
		LOGGER.info(">> terminateThreads " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<ModulesResponse> modules(ModulesArguments args) {
		LOGGER.info(">> modules " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<LoadedSourcesResponse> loadedSources(LoadedSourcesArguments args) {
		LOGGER.info(">> loadedSources " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<StepInTargetsResponse> stepInTargets(StepInTargetsArguments args) {
		LOGGER.info(">> stepInTargets " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<GotoTargetsResponse> gotoTargets(GotoTargetsArguments args) {
		LOGGER.info(">> gotoTargets " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<CompletionsResponse> completions(CompletionsArguments args) {
		LOGGER.info(">> completions " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<ExceptionInfoResponse> exceptionInfo(ExceptionInfoArguments args) {
		LOGGER.info(">> exceptionInfo " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<ReadMemoryResponse> readMemory(ReadMemoryArguments args) {
		LOGGER.info(">> readMemory " + args);
		throw new UnsupportedOperationException();
	}

	@Override
	public CompletableFuture<DisassembleResponse> disassemble(DisassembleArguments args) {
		LOGGER.info(">> disassemble " + args);
		throw new UnsupportedOperationException();
	}

	protected void clientOutput(String output, String category) {
		if (this.client != null) {
			OutputEventArguments arguments = new OutputEventArguments();
			arguments.setCategory(category);
			arguments.setOutput(output);
			this.client.output(arguments);
		}
	}

	protected void clientErr(String output) {
		clientOutput(output, OutputEventArgumentsCategory.STDERR);
	}

	protected void clientOut(String output) {
		clientOutput(output, OutputEventArgumentsCategory.STDOUT);
	}

}
